/* -*-mode:C;c-basic-offset:8-*-
 *      Copyright (C) 1993-1995 Bas Laarhoven.
 *
 * Considerable changes by Claus-Justus Heine, 1994, 1995
 * implemented compression and random block access
 *

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 $Source: /homes/cvs/zftape/ftape-read.c,v $
 $Author: claus $
 *
 $Revision: 1.22 $
 $Date: 1996/08/12 10:28:56 $
 $State: Exp $
 *
 *      This file contains the reading code
 *      for the QIC-117 floppy-tape driver for Linux.
 */

static char RCSid[] =
"$Id: ftape-read.c,v 1.22 1996/08/12 10:28:56 claus Exp $";


#include "ftape.h"
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <asm/segment.h>

#include "qic80-compress.h"
#include "ftape-rw.h"
#include "ftape-ctl.h"
#include "ecc.h"
#include "ftape-read.h"
#include "ftape-write.h"
#include "ftape-io.h"
#include "qic117.h"
#include "ftape-eof.h"
#include "qic80-vtbl.h"
#include "ftape-bsm.h"

/*      Global vars.
 */
int just_before_eof = 0;
int buf_len_rd = 0;

/*      Local vars.
 */

void
ftape_zap_read_buffers( void)
{
	TRACE_FUN( 4, "ftape_zap_read_buffers");
	int i;
	
	zftape_init_tape_buffer();
	for (i = 0; i < ftape_num_buffers; ++i) {
		buffer[ i].status = waiting;
		buffer[ i].bytes = 0;
		buffer[ i].skip = 0;
		buffer[ i].retry = 0;
	}
	buf_len_rd = 0;
	ftape_state = idle;
	TRACE_EXIT;
}

static SectorMap
convert_sector_map( buffer_struct* buff)
{
	TRACE_FUN( 5, "convert_sector_map");
	int i = 0;
	SectorMap bad_map = get_bad_sector_entry( buff->segment_id);
	SectorMap src_map = buff->soft_error_map | buff->hard_error_map;
	SectorMap dst_map = 0;

	if (bad_map || src_map) {
		TRACEx1( 7, "bad_map = 0x%08x", (unsigned int)bad_map);
		TRACEx1( 7, "src_map = 0x%08x", (unsigned int)src_map);
	}
	while (bad_map) {
		while ((bad_map & 1) == 0) {
			if (src_map & 1) {
				dst_map |= (1 << i);
			}
			src_map >>= 1;
			bad_map >>= 1;
			++i;
		}
		/* (bad_map & 1) == 1 */
		src_map >>= 1;
		bad_map >>= 1;
	}
	if (src_map) {
		dst_map |= (src_map << i);
	}
	if (dst_map) {
		TRACEx1( 7, "dst_map = 0x%08x", (unsigned int)dst_map);
	}
	TRACE_EXIT;
	return dst_map;
}

static int
correct_and_copy_fraction( unsigned int tail, byte* destination, 
			   int start, int size)
{
	TRACE_FUN( 5, "correct_and_copy_fraction");
	struct memory_segment mseg;
	int result;
	SectorMap read_bad;
	
	mseg.read_bad = convert_sector_map( &buffer[ tail]);
	mseg.marked_bad = 0;          /* not used... */
	mseg.blocks = buffer[ tail].bytes / SECTOR_SIZE;
	mseg.data = buffer[ tail].address;
	/*    If there are no data sectors we can skip this segment.
	 */
	if (mseg.blocks <= 3) {
		TRACE( 7, "empty segment");
		TRACE_EXIT;
		return 0;
	}
	read_bad = mseg.read_bad;
	history.crc_errors += count_ones( read_bad);
	result = ecc_correct_data( &mseg);
	if (read_bad != 0 || mseg.corrected != 0) {
		TRACEx( 4, "crc error map:", (unsigned int)read_bad);
		TRACEx( 4, "corrected map:", (unsigned int)mseg.corrected);
		history.corrected += count_ones( mseg.corrected);
	}
	if (result == ECC_CORRECTED || result == ECC_OK) {
		if (result == ECC_CORRECTED) {
			TRACEi( 4, "ecc corrected segment:", 
				buffer[ tail].segment_id);
		}
		if( start < 0 ) {
			start= 0;
		}
		if( (start+size) > ((mseg.blocks - 3) * SECTOR_SIZE) ) {
			size = (mseg.blocks - 3) * SECTOR_SIZE  - start;
		} else if ( size < 0 ) {
			size= 0;
		}
		if( size > 0 ) {
			memcpy( destination + start, mseg.data + start, size );
		}
		if ((read_bad ^ mseg.corrected) & mseg.corrected) {
			/* sectors corrected without crc errors set */
			history.crc_failures++;
		}
		TRACE_EXIT;
		return size;
	} else {
		TRACEi( 1, "ecc failure on segment", buffer[ tail].segment_id);
		history.ecc_failures++;
		TRACE_EXIT;
		return -EAGAIN;             /* should retry */
	}
	TRACE_EXIT;
	return 0;
}

/*      Read given segment into buffer at address.
 */
int
read_segment_fraction( unsigned segment_id, byte* address, int read_ahead,
                       int start, int  size)
{
	TRACE_FUN( 5, "read_segment");
	int read_done= 0;
	int result = 0;
	int bytes_read = 0;
	int retry = 0;
	
	TRACEi( 7, "segment_id =", segment_id);
	if (ftape_state != reading) {
		if (ftape_state == writing) {
			TRACE( 3, "Can this ever happen?");
			ftape_flush_buffers( 0 ); /* can this ever happen? */
			TRACE( 7, "calling ftape_abort_operation");
			result = ftape_abort_operation();
			if (result < 0) {
				TRACE( 1, "ftape_abort_operation failed");
				TRACE_EXIT;
				return -EIO;
			}
		} else {
			/* clear remaining read buffers */
			ftape_zap_read_buffers();
		}
		ftape_state = reading;
	}
	if (segment_id >= segments_per_track * tracks_per_tape) {
		TRACE( 4, "reading past end of tape");
		TRACE_EXIT;
		return -ENOSPC;
	}
	for (;;) {
		/* Search all full buffers for the first matching the wanted 
		 * segment. Clear other buffers on the fly.
		 */
		while (!read_done && buffer[ tail].status == done) {
			if (buffer[ tail].segment_id == segment_id) {
				/* If out buffer is already full, 
				 * return its contents.
				 */
				if (buffer[ tail].deleted) {
					TRACE_EXIT;
		/*  Return a value that read_header_segment understands.
		 *  As this should only occur when searching for the header
		 *  segments it shouldn't be misinterpreted elsewhere.
		 */
					return 0;
				}
				TRACEi( 5, "found segment in cache :", 
					segment_id);
				result= correct_and_copy_fraction(tail,address,
								  start,size);
				TRACEi( 7, "segment contains (bytes) :", 
					result);
				if (result < 0) {
					if (result != -EAGAIN) {
						TRACE_EXIT;
						return result;
					}
					/* keep read_done == 0, will trigger 
					 * ftape_abort_operation because 
					 * reading wrong segment.
					 */
					TRACE( 2, "ecc failed, retry");
					++retry;
				} else {
					read_done = 1;
				}
				bytes_read = result;
			} else {
				TRACEi( 7, "zapping segment in cache :", 
					buffer[ tail].segment_id);
			}
			buffer[ tail].status = waiting;
			next_buffer( &tail);
		}
		if (!read_done && buffer[ tail].status == reading) {
			if (buffer[ tail].segment_id == segment_id) {
				int result = wait_segment( reading);
				if (result < 0) {
					if (result == -EINTR) {
						TRACE_EXIT;
						return result;
					}
					TRACE( 1, "wait_segment failed "
					       "while reading");
					ftape_abort_operation();
				}
			} else {
				/*We're reading the wrong segment, stop runner.
				 */
				TRACE( 5, "reading wrong segment, "
				       "abort operation");
				ftape_abort_operation();
			}
		}
		/* if just passed the last segment on a track, wait for BOT
		 * or EOT mark.
		 */
		if (runner_status == logical_eot) {
			int status;
			result = ftape_ready_wait( timeout.seek, &status);
			if (result < 0) {
				TRACE( 1, "ftape_ready_wait waiting for "
				       "eot/bot failed");
			}
			if ( (status & (QIC_STATUS_AT_BOT | QIC_STATUS_AT_EOT))
			     == 
			     0) {
				TRACE( 1, "eot/bot not reached");
			}
			runner_status = end_of_tape;
		}
		/*    should runner stop ?
		 */
		if (runner_status == aborting || 
		    runner_status == buffer_overrun ||
		    runner_status == end_of_tape) {
			if (runner_status != end_of_tape &&
			    !(runner_status == aborting && !tape_running)) {
				ftape_dumb_stop();
			}
			if (runner_status == aborting) {
				if (buffer[ head].status == reading || 
				    buffer[ head].status == error) {
					if (buffer[ head].status == error) {
		history.defects += count_ones( buffer[ head].hard_error_map);
					}
					buffer[ head].status = waiting;
				}
			}
			runner_status = idle;       /* aborted ? */
		}
		/*    If segment to read is empty, do not start runner for it,
		 *    but wait for next read call.
		 */
		if (get_bad_sector_entry( segment_id) == EMPTY_SEGMENT) {
			bytes_read = 0;           /* flag empty segment */
			read_done = 1;
		}
		/*  Allow escape from this loop on signal !
		 */
		if (current->signal & _DONT_BLOCK) {
			TRACE( 2, "interrupted by non-blockable signal");
			TRACE_EXIT;
			return -EINTR;
		}
		/*    If we got a segment: quit, or else retry up to limit.
		 */
		if (read_done) {
			break;
		}
		if (retry > RETRIES_ON_ECC_ERROR) {
			history.defects++;
			TRACE( 1, "too many retries on ecc failure");
			TRACE_EXIT;
			return -ENODATA;
		}
		/*    Now at least one buffer is empty !
		 *    Restart runner & tape if needed.
		 */
		TRACEx3( 8, "head: %d, tail: %d, runner_status: %d",
			 head, tail, runner_status);
		TRACEx2( 8, "buffer[].status, [head]: %d, [tail]: %d",
			 buffer[ head].status, buffer[ tail].status);
		if (buffer[ tail].status == waiting) {
			setup_new_segment( &buffer[ head], segment_id, -1);
			if (!read_ahead) {
				buffer[ head].next_segment = 0; /* disable 
								 * read-ahead 
								 */
			}
			calc_next_cluster( &buffer[ head]);
			if (runner_status == idle) {
				result = ftape_start_tape( 
					  segment_id,
					  buffer[ head].sector_offset);
				if (result < 0) {
					TRACEx1( 1, "segment %d unreachable", 
						 segment_id);
					TRACE_EXIT;
					return result;
				}
				runner_status = running;
			}
			buffer[ head].status = reading;
			setup_fdc_and_dma( &buffer[ head], FDC_READ);
		}
	}
	if (read_done) {
		TRACE_EXIT;
		return bytes_read;
	} else {
		TRACE( 1, "too many retries");
		TRACE_EXIT;
		return -EIO;
	}
}

static int
calc_tape_len(  format_type format_code, int *tape_len ) 
{
	TRACE_FUN( 5, "calc_tape_len");
	int new_tape_len = *tape_len;

	switch (format_code) {
	case fmt_425ft:
		new_tape_len = 425;
		break;
	case fmt_normal:
		if (*tape_len == 0) {        /* otherwise 307 ft */
			new_tape_len = 205;
		}
		break;
	case fmt_1100ft:
		new_tape_len = 1100;
		break;
	case fmt_wide: {
		int segments_per_1000_inch = 1; /* non-zero default for switch */
		switch (qic_std) {
		case QIC_TAPE_QIC40:
			segments_per_1000_inch = 332;
			break;
		case QIC_TAPE_QIC80:
			segments_per_1000_inch = 488;
			break;
		case QIC_TAPE_QIC3010:
			segments_per_1000_inch = 730;
			break;
		case QIC_TAPE_QIC3020:
			segments_per_1000_inch = 1430;
			break;
		}
		new_tape_len = (   1000 * segments_per_track 
				 + (segments_per_1000_inch - 1)) 
			       / segments_per_1000_inch;
		break;
	}
	default:
		TRACE( 1, "unknown tape format, please report !");
		TRACE_EXIT;
		return -EIO;
	}
	if (new_tape_len != *tape_len) {
		*tape_len = new_tape_len;
		TRACEx1( 4, "calculated tape length is %d ft", *tape_len);
		ftape_calc_timeouts();
	}
	TRACE_EXIT;
	return 0;
}


static int
set_tape_parms( unsigned int max_floppy_side, unsigned int max_floppy_track, 
		  unsigned int max_floppy_sector )
{
	TRACE_FUN( 5, "set_floppy_parms");
	
	if (segments_per_track == 0 && tracks_per_tape == 0 &&
	    max_floppy_side == 0 && max_floppy_track == 0 &&
	    max_floppy_sector == 0) {
		/*  QIC-40 Rev E and earlier has no values in the header.
		 */
		segments_per_track = 68; 
		tracks_per_tape = 20;
		max_floppy_side = 1;
		max_floppy_track = 169;
		max_floppy_sector = 128;
	}
	/*  This test will compensate for the wrong parameter on tapes
	 *  formatted by Conner software.
	 */
	if (segments_per_track == 150 &&
	    tracks_per_tape == 28 &&
	    max_floppy_side == 7 &&
	    max_floppy_track == 149 &&
	    max_floppy_sector == 128) {
		TRACE( 2, "the famous CONNER bug: "
		       "max_floppy_side off by one !");
		max_floppy_side = 6;
	}
	/*  This test will compensate for the wrong parameter on tapes
	 *  formatted by Colorado Windows software.
	 */
	if (segments_per_track == 150 &&
	    tracks_per_tape == 28 &&
	    max_floppy_side == 6 &&
	    max_floppy_track == 150 &&
	    max_floppy_sector == 128) {
		TRACE( 2, "the famous Colorado bug: "
		       "max_floppy_track off by one !");
		max_floppy_track = 149;
	}
	segments_per_head = ((max_floppy_sector / SECTORS_PER_SEGMENT) *
			     (max_floppy_track + 1));
	/*
	 *    Verify drive_configuration with tape parameters
	 */
	if ( segments_per_head == 0 || segments_per_cylinder == 0 ||
	     ((segments_per_track * tracks_per_tape - 1) / segments_per_head
	      != 
	      max_floppy_side) ||
	     (segments_per_head / segments_per_cylinder - 1 
	      != max_floppy_track) ||
	     (segments_per_cylinder * SECTORS_PER_SEGMENT != max_floppy_sector)
#ifdef TESTING
	     || (format_code == 4 && 
		 (max_floppy_track != 254 || max_floppy_sector != 128))
#endif
	     ) {
		TRACE( 1, "Tape parameters inconsistency, please report");
		TRACE_EXIT;
		return -EIO;
	}
	TRACE_EXIT;
	return 0;
}


static int
read_table_segments( byte *address )      
{
	TRACE_FUN( 6, "read_table_segments");
	int result;
	int i;

	for ( i = used_header_segment+1; i < first_data_segment; i++) {
		(void)read_segment( i, address, 1); /*keep streaming */
	}
	result = read_segment( first_data_segment, address, 1);
	if( result < 0 ) {
		if (result == -EINTR) {
			TRACE_EXIT;
			return result;
		}
		TRACEi( 2, "Couldn't read volume table segment at", 
			first_data_segment );
		memset( address, '\0', SIZEOF(deblock_buffer) );
	} 
	if ( qic80_extract_volume_headers( address ) < 0 ) {
		first_data_segment = -1;
		TRACE_EXIT;
		return -ENOMEM;
	}
	result = zftape_extract_compression_map( address, 
						 zftape_compression_map_location );
	if ( result < 0 ) {
		switch ( result ) {
		case -EINTR: {
			TRACE_EXIT;
			return result;
		}
		case -ENOMEM: {
			first_data_segment = -1;
			TRACE( 1, "unsufficient memory for "
			       "compression map");
			TRACE_EXIT;
			return -ENOMEM;
		}
		default: {
			TRACEi( 2, "Error extracting "
				"compression map,", result);
			zftape_compression_map_location = 0;
			TRACE_EXIT;
			return 0; /* don't worry */
		}
		}
	}
	TRACE_EXIT;
	return 0;
}

int
read_header_segment( byte* address, int header_only )
{
	TRACE_FUN( 5, "read_header_segment");
	int result;
	int header_segment = -1;
	int first_failed = 0;
	unsigned int max_floppy_side, max_floppy_track, max_floppy_sector;
	int status;
	
	result = ftape_report_drive_status( &status);
	if (result < 0) {
		TRACE( 1, "error: error_status or report failure");
		TRACE_EXIT;
		return -EIO;
	}
	TRACE( 7, "reading...");
	ftape_last_data_segment = 68;   /* will allow us to read the header ! */
	/*  We're looking for the first header segment.
	 *  A header segment cannot contain bad sectors, therefor at the
	 *  tape start, segments with bad sectors are (according to QIC-40/80)
	 *  written with deleted data marks and must be skipped.
	 */
	used_header_segment = -1;
	result = 0;
	for (header_segment = 0;
	     header_segment < ftape_last_data_segment && result == 0;
	     ++header_segment) {
		/*  Set no read-ahead, the isr will force read-ahead whenever
		 *  it encounters deleted data !
		 */
		result = read_segment( header_segment, address, !header_only );
		if (result < 0 && !first_failed) {
			TRACE( 1, "header segment damaged, trying backup");
			first_failed = 1;
			result = 0;               /* force read of next (backup) segment */
		}
	}
	if (result < 0 || header_segment >= ftape_last_data_segment) {
		TRACE( 1, "no readable header segment found");
		TRACE_EXIT;
		return -EIO;
	}
	if (GET4( address, 0) != 0xaa55aa55) {
		TRACE( 1, "wrong signature in header segment");
		TRACE_EXIT;
		return -EIO;
	}
	header_segment_1 = GET2( address, 6);
	header_segment_2 = GET2( address, 8);
	TRACEx2( 4, "header segments are %d and %d",
		 header_segment_1, header_segment_2);
	used_header_segment = (first_failed) ? header_segment_2 : header_segment_1;
	
	/*    Verify tape parameters...
	 *    QIC-40/80 spec:                 tape_parameters:
	 *
	 *    segments-per-track              segments_per_track
	 *    tracks-per-cartridge            tracks_per_tape
	 *    max-floppy-side                 (segments_per_track *
	 *                                    tracks_per_tape - 1) /
	 *                                    segments_per_head
	 *    max-floppy-track                segments_per_head /
	 *                                    segments_per_cylinder - 1
	 *    max-floppy-sector               segments_per_cylinder *
	 *                                    SECTORS_PER_SEGMENT
	 */
	format_code = (format_type) *(address + 4);
	segments_per_track = GET2( address, 24);
	tracks_per_tape = *(address + 26);
	max_floppy_side = *(address + 27);
	max_floppy_track = *(address + 28);
	max_floppy_sector = *(address + 29);
	TRACEx6( 4, "(fmt/spt/tpc/fhm/ftm/fsm) = %d/%d/%d/%d/%d/%d",
		 format_code, segments_per_track, tracks_per_tape,
		 max_floppy_side, max_floppy_track, max_floppy_sector);

	result = calc_tape_len( format_code, &tape_len ) ;
	if ( result < 0 ) {
		TRACE_EXIT;
		return result;
	}
	result = set_tape_parms( max_floppy_side, max_floppy_track, max_floppy_sector );
	if ( result < 0 ) {
		TRACE_EXIT;
		return result;
	}

	first_data_segment = GET2( address, 10); /* first data segment */
	ftape_last_data_segment = GET2( address, 12);
	TRACEi( 4, "first data segment:", first_data_segment);
	TRACEi( 4, "last data segment:", ftape_last_data_segment);
	if ( extract_bad_sector_map( address) < 0  ) {
		first_data_segment = -1;
		TRACE_EXIT;
		return -ENOMEM;
	}
	strncpy(ftape_label,&address[30],44);
	ftape_old_ftape = ftape_validate_label( ftape_label ) != 0;
	ftape_extract_file_marks( address);
	ftape_calc_data_amount();
	if ( ftape_old_ftape ) {
		TRACE( 2, "Found old ftaped tape, emulating eof marks, "
		       "entering read-only mode");
		if (qic80_fake_volume_headers(eof_map, nr_of_eof_marks) < 0 ) {
			first_data_segment = -1;
			TRACE_EXIT;
			return -ENOMEM;
		}
	} else {
		/* for simplicity */
#if 0
		zftape_eliminate_failed_sector_log(  eof_map, nr_of_eof_marks);
#endif
		nr_of_eof_marks = 0; 
		zftape_compression_map_location = GET2( address, 78);
		if ( !header_only ) {
			result = read_table_segments( address );
		}
	}
	zftape_set_flags( ftape_unit ); /* sets raw mode and compression flag*/
	ftape_reset_position();
	TRACE_EXIT;
	return result;
}

/*
 * out:
 *
 * int *read_cnt: the number of bytes we removed from the deblock_buffer
 *                (result)
 * int *to_do   : the remaining size of the read-request. Is changed.
 *
 * int:
 *
 * char *buff      : buff is the address of the upper part of the user buffer,
 *                   that hasn't been filled with data yet.
 * int buf_pos_read: copy of buf_pos_rd
 * int buf_len_read: copy of buf_len_rd
 * char *deblock_buffer: ftape_deblock_buffer
 *
 * returns the amount of data actually copied to the user-buffer
 *
 * *to_do MUST NOT SHRINK except to indicate an EOT. In this case *to_do has to
 * set to 0
 * We cannot return -ENOSPC, because we return the amount of data actually
 * copied to the user-buffer
 *
 */
static int
ftape_simple_read ( int *read_cnt,
                    int to_do,
                    char *buff,
                    int buf_pos_read,
                    int segment_size,
                    char *deblock_buffer,
                    unsigned short dummy )
{
	TRACE_FUN( 5, "ftape_simple_read");
	
	*read_cnt = (buf_len_rd < to_do) ? buf_len_rd : to_do;
	memcpy_tofs( buff, deblock_buffer + buf_pos_read, *read_cnt);
	ftape_uncmpr_pos += *read_cnt;
	TRACEi( 7, "nr bytes just read:", *read_cnt);
	TRACE_EXIT;
	return *read_cnt;
}


static int
check_read_access( int *req_len, int *req_clipped, 
		   int *block_size, int *vol_no )
{
	TRACE_FUN( 6, "check_read_access");
	int result = 0;
	static int remaining = 0;
	
	if ( ftape_offline || !formatted || no_tape) {
		TRACEx3( 1, "offline = %d, formatted = %d, no_tape = %d",
			 ftape_offline, formatted, no_tape);
		ftape_io_state = io_idle;
		TRACE_EXIT;
		return -EIO;
	} 
	history.used |= 1;
	if ( (first_data_segment == -1) &&
	     ((result = read_header_segment( deblock_buffer, 0 )) < 0) ) {
		ftape_io_state = io_idle;
		TRACE_EXIT;
		return result;
	}
	if ( ftape_seg_pos > ftape_last_data_segment ) {
		TRACE_EXIT;
		return -ENOSPC;
	} else if ( qic80_check_for_eom( ftape_seg_pos ) ) {
		TRACE_EXIT;
		return -ENODATA;
	}
	if ( ftape_io_state != io_reading ) {
		*vol_no = qic80_find_file_no ( ftape_seg_pos );
		*block_size = qic80_get_volume_block_size ( *vol_no );
	 	if ( (*req_len % *block_size) != 0 ) {
			TRACE_EXIT;
			return -EINVAL;
		}
		/* check for file mark or end of tape if using raw mode
		 * remaining will be negative if we exceed the volume size.
		 */
		remaining = qic80_check_for_eof( *vol_no, 
						 ftape_uncmpr_pos );
		ftape_io_state = io_reading;
		TRACEx2( 5, "remaining: %d, vol_no: %d", remaining, *vol_no );
	} else 	if ( (*req_len % *block_size) != 0 ) {
		TRACE_EXIT;
		return -EINVAL;
	}
	remaining -= *req_len;
	/* As GNU tar doesn't accept partial read counts when the multiple
	 * volume flag is set, we make sure to return the requested amount
	 * of data. Except, of course, at the end of the tape or file mark.
	 */
	if ( remaining <= 0 ) {
		TRACEx2( 7, "clipped request from %d to %d.", 
			    *req_len, (int)(*req_len + remaining) );
		*req_len += remaining;
		*req_clipped = 1;
	} else {
		*req_clipped = 0;
	}
	TRACE_EXIT;
	return result;
}

static int
read_deblock_buffer_from_tape( int *this_segs_size )
{
	TRACE_FUN( 6, "read_deblock_buffer_from_tape");

	while ((*this_segs_size = ftape_get_seg_size( ftape_seg_pos )) == 0 ) {
		ftape_seg_pos ++;
		ftape_seg_data_pos = 0;
	}
	if ( ftape_seg_pos != ftape_deblock_segment ) {
		buf_len_rd = read_segment( ftape_seg_pos, deblock_buffer,  1 );
		if ( buf_len_rd != *this_segs_size ) {
			TRACEx1( 7, "read_segment result: %d", buf_len_rd);
			TRACEx3( 1, "Segment %d should contain "
				 "%d bytes, but we got only %d",
				 ftape_seg_pos, *this_segs_size, buf_len_rd );
			ftape_deblock_segment = -1;
			TRACE_EXIT;
			return buf_len_rd >= 0 ? -EIO : buf_len_rd;
		};
		ftape_deblock_segment = ftape_seg_pos;
	} 
	buf_len_rd = *this_segs_size - ftape_seg_data_pos;
	TRACE_EXIT;
	return 0;
}

static int
empty_deblock_buffer( int this_segs_size, char *buff, int req_len, 
		      int block_size, int vol_no )
{
	TRACE_FUN( 6, "empty_deblock_buffer" );
	int cnt;
	int result = 0;

	TRACEi( 7, "this_segs_size: ", this_segs_size );
	if ( zftape_use_compression 
	     && qic80_get_volume_cmpr_flag( vol_no ) ) {
		result= zftape_compress_read ( &cnt, req_len, buff,
					       ftape_seg_data_pos, 
					       this_segs_size,
					       deblock_buffer, block_size );
	} else {                                  
		result= ftape_simple_read ( &cnt, req_len, buff,
					    ftape_seg_data_pos, 
					    this_segs_size,
					    deblock_buffer, block_size );
	}
	if (result < 0 ) {
		TRACEi( 1, "ftape_read_fun() failed with result ", result );
		TRACE_EXIT;
		return result;
	}
	ftape_data_pos     += cnt;
	ftape_seg_data_pos += cnt;
	buf_len_rd         -= cnt; /* remaining bytes in buffer */
	TRACEx2( 7, "buf_len_rd: %d, cnt: %d", buf_len_rd, cnt );
	if( ftape_seg_data_pos >= this_segs_size ) {
		ftape_seg_pos++;
		ftape_seg_data_pos = 0;
	}

	TRACEi( 7, "bytes moved out of deblock-buffer: ", cnt );
	TRACE_EXIT;
	return result;
}


/* 
 * note: we store the segment id of the segment that is inside the
 * deblock buffer. This spares a lot of read_segment()s when we use
 * small block-sizes. The block-size may be 1kb (SECTOR_SIZE). In this
 * case a MTFSR 28 maybe still inside the same segment.
 */
int
_ftape_read( char* buff, int req_len )
{
	TRACE_FUN( 5, "_ftape_read");
	int result = 0;
	int bytes_read = 0;
	static int this_segs_size = 0;
	int req_clipped;
	static int block_size = 1;
	static int vol_no     = 0;
	
	ftape_resid = req_len;
	
	if ( (result = verify_area( VERIFY_WRITE, buff, req_len )) != 0 ) {
		TRACE( 1, "verify_area failed");
		TRACE_EXIT;
		return result;
	}
	if ( (result = check_read_access( &req_len, &req_clipped, 
					  &block_size, &vol_no ) ) < 0 ) {
		TRACE_EXIT;
		return result;
	}
	/* don't return with a partial count ! 
	 */
	while (req_len > 0) { 
		/* buf_len_rd == 0 means that we need to read a new segment.
		 * in that case we first check ftape_seg_pos and sector_pos
		 * to be consistent and then compute this_segs_size
		 */
		if ( buf_len_rd == 0 &&
		     (result = 
		      read_deblock_buffer_from_tape( &this_segs_size )) < 0 ) {
			ftape_resid -= bytes_read;
			TRACE_EXIT;
			return result;
		}
		/*  Allow escape from this loop on signal !
		 */
		if (current->signal & _DONT_BLOCK) {
			TRACE( 2, "interrupted by non-blockable signal");
			ftape_resid -= bytes_read;
			TRACE_EXIT;
			return -EINTR;
		}
		if ( (result =  empty_deblock_buffer( this_segs_size, buff, 
						      req_len, 
						      block_size, 
						      vol_no ) ) < 0 ) {
			ftape_resid -= bytes_read;
			TRACE_EXIT;
			return result;
		}
		TRACEi( 7, "bytes just read: ", result );
		bytes_read += result; /* what we got so far       */
		buff       += result; /* index in user-buffer     */
		req_len    -= result; /* what's left from req_len */
	} /* while ( req_len  > 0 ) */
	if ( req_clipped ) {
		TRACE( 7, "maybe partial count because of eof mark");
		if ( just_before_eof && bytes_read == 0 ) {
			/*  req_len was > 0, but user didn't get anything
			 *  the user has read in the eof-mark
			 */
			just_before_eof = 0;
			ftape_zap_read_buffers();
			zftape_advance_to_next_segment();
			ftape_uncmpr_pos = 0;
			zftape_reset_cmpr_locals(); /* sure is sure */
		} else {
			/* don't skip to the next file before the user
			 * tried to read a second time past EOF
			 * Just mark that we are at EOF and 
			 * maybe decrement ftape_seg_pos to stay in the same
			 * volume;
			 */
			just_before_eof = 1;
			zftape_retard_to_prev_segment();
			TRACE( 4, "just before eof");
		}
	}
	ftape_resid -= result; /* for MTSTATUS       */
	TRACE_EXIT;
	return bytes_read;
}







